/*
     This file is part of GNUnet.
     (C) 2001, 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @author Igor Wronsky
 * @file applications/afs/database/high_mysql.c
 *
 * Database: MySQL
 *
 * NOTE: This db module does NOT work with mysql v3.23.49 due
 * to a bug in mysql. Use a newer version. The only version that
 * we know it works on, is the version used for development,
 * v3.23.56, though any later version is probably also fine.
 *
 * The main purpose of mysql support was to gauge if mysql scales better 
 * and/or has fewer catastrophic failures than gdbm or tdb. Though no
 * actual benchmarks have been done, this module hasn't caused any 
 * trouble (to the author) and has given a feeling of better
 * performance.
 *
 * Using this module requires some manual setup from the MySQL side. 
 * Instructions follow.
 *
 * To use: in gnunet.conf, set
 *
 *  DATABASETYPE = mysql
 *  
 * Then with mysql and root priviledges, do the following. [You should
 * replace $USER with the username thats running the gnunetd process].
 *
    CREATE DATABASE gnunet;
    GRANT select,insert,update,delete,create,alter,drop
       ON gnunet.* TO $USER;
    SET PASSWORD FOR $USER=PASSWORD('$your_new_password');
    FLUSH PRIVILEGES;
 *
 * Then write a $HOME/.my.cnf file having the following
 
    [client]
    user=$USER
    password=$your_new_password
    
 * Thats all. Note that .my.cnf file is a security risk unless
 * its on a safe partition etc. Even greater security risk can
 * be achieved by setting no password to $USER. Luckily
 * your $USER has only priviledges to mess up gnunet's tables,
 * nothing else (unless you give him more, of course).
 *
 * For MySQL running on standard ext2/ext3 fs, this results in a db 
 * thats supposed to hold two or four gigs. The only option to
 * extend this (for a single bucket solution) is to put the table
 * on some filesystem that allows large files. See MySQL manual
 * for more info. The MySQL RAID table type doesn't seem to work here.
 *
 * Todo:
 * - document better how to setup the mysql database 
 * - After 16384 search results per key, mediumblob gets 
 *   saturated (note that returning 16384 search results of 1k
 *   is already a slight memory usage peak - for any dbmgr!).
 * - What is getRandomContent() actually supposed to return in ce? 
 *   And what kind of content should be matched... -IW 
 *
 * 
 **/

#include "config.h"
#include "applications/afs/database/high_backend.h"
#include "util/checksum.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/semaphore.h"
#include "util/statuscalls.h"
#include "util/storage.h"
#include "util/xmalloc.h"

#include <math.h>
#include <netinet/in.h>
#include <mysql/mysql.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>
#include <errno.h>

/**
 * A mysql wrapper
 **/
typedef struct {
  MYSQL * dbf;
  unsigned int i;
  unsigned int n;
  Mutex DATABASE_Lock_;
} mysqlHandle;

HighDBHandle initContentDatabase(unsigned int i,
				 unsigned int n) {
  mysqlHandle * dbh;
  char * scratch;

  dbh = MALLOC(sizeof(mysqlHandle));
  dbh->dbf = mysql_init(NULL);
  dbh->i = i;
  dbh->n = n;
  if(dbh->dbf == NULL) {
    LOG(LOG_ERROR, 
	"ERROR: Unable to init MySQL\n");
    FREE(dbh);
    return NULL;
  }

  mysql_options(dbh->dbf, 
		MYSQL_READ_DEFAULT_GROUP, 
		"client");
  mysql_real_connect(dbh->dbf,
		     NULL,
		     NULL,
		     NULL,
		     "gnunet",
		     0,
		     NULL,
		     0);
  if (mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR: %s\n", 
	mysql_error(dbh->dbf));
    FREE(dbh);
    return NULL;
  }    

  scratch = MALLOC(1024);
  sprintf(scratch,
  	  "CREATE TABLE IF NOT EXISTS data%uof%u ("
	  "  hash tinyblob NOT NULL default'',"
	  "  priority int(11) NOT NULL default 0,"
	  "  type tinyint NOT NULL default 0,"
	  "  fileIndex smallint NOT NULL default 0,"
	  "  fileOffset int(11) NOT NULL default 0,"
	  "  doubleHash tinyblob NOT NULL default '',"
	  "  content mediumblob NOT NULL default '',"
	  "  PRIMARY KEY (hash(20)),"
	  "  KEY priority (priority)"
	  ") TYPE=MyISAM",
	  dbh->n,
	  dbh->i);
  mysql_query(dbh->dbf,
  	      scratch);
  FREE(scratch);
  if (mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR: %s\n", 
	mysql_error(dbh->dbf));
    FREE(dbh);
    return NULL;
  }
  MUTEX_CREATE_RECURSIVE(&dbh->DATABASE_Lock_);	

  return dbh;
}

/**
 * Clean shutdown of the storage module
 **/
void doneContentDatabase(HighDBHandle handle) {
  mysqlHandle * dbh = handle;
 
  MUTEX_DESTROY(&dbh->DATABASE_Lock_);
  mysql_close(dbh->dbf);
  FREE(dbh);
}


/**
 * Call a method for each key in the database and
 * call the callback method on it. 
 *
 * @param callback the callback method
 * @param data second argument to all callback calls
 * @return the number of items stored in the content database
 **/
int forEachEntryInDatabase(HighDBHandle handle,
 		           EntryCallback callback,
	    	           void * data) {
  mysqlHandle * dbh = handle;
  MYSQL_RES *sql_res;
  MYSQL_ROW sql_row;
  ContentIndex ce;
  void * result;
  char * scratch;
  int count = 0;
  int len;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  scratch = MALLOC(256);
  sprintf(scratch,
  	  "SELECT content,type,priority,doubleHash,fileOffset,fileIndex,hash FROM data%uof%u",
	  dbh->n,
	  dbh->i);
  mysql_query(dbh->dbf, 
	      scratch);
  FREE(scratch);
  if (mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR: %s\n", 
	mysql_error(dbh->dbf));
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return(SYSERR);
  }
  
  if (!(sql_res=mysql_use_result(dbh->dbf))) {
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return(SYSERR);
  }

  while ((sql_row=mysql_fetch_row(sql_res))) {   
    len = mysql_fetch_lengths(sql_res)[0];
    if (len > 0) {
      result = MALLOC(len);
      memcpy(result,
	     sql_row[0],
	     len);
    } else
      result = NULL;
    
    ce.type = htons(atol(sql_row[1]));
    ce.importance = htonl(atol(sql_row[2]));
    if (ntohs(ce.type)==LOOKUP_TYPE_3HASH) {
      if (mysql_fetch_lengths(sql_res)[3] == sizeof(HashCode160))
	memcpy(&ce.hash, 
	       sql_row[3],
	       sizeof(HashCode160));
    } else {
      memcpy(&ce.hash, 
	     sql_row[6],
	     sizeof(HashCode160));
    }
    ce.fileOffset = htonl(atol(sql_row[4]));
    ce.fileNameIndex = htons(atol(sql_row[5]));       
    callback((HashCode160*)sql_row[6],
	     &ce,
	     result, /* freed by callback */
	     len,
	     data);
    count++;
  }
		
  mysql_free_result(sql_res);
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return count;
}

/**
 * Get the number of entries in the database.
 **/
int countContentEntries(HighDBHandle handle) {
  mysqlHandle * dbh = handle;
  MYSQL_RES * sql_res;
  MYSQL_ROW sql_row;
  char * scratch;
  int count=0;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  scratch = MALLOC(128);
  sprintf(scratch, 
          "SELECT count(*) FROM data%uof%u",
	  dbh->n,
  	  dbh->i);
  mysql_query(dbh->dbf,
  	      scratch);
  FREE(scratch);
  if (!(sql_res=mysql_store_result(dbh->dbf))) {
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return(SYSERR);
  }

  if ((sql_row=mysql_fetch_row(sql_res))) {
    count = atol(sql_row[0]);
  }
  mysql_free_result(sql_res);
     
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return count;
}

/**
 * Read the contents of a bucket to a buffer.
 *
 * @param fn the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, -1 on failure
 **/ 
int readContent(HighDBHandle handle,
		HashCode160 * query,
	        ContentIndex * ce,
	        void ** result,
		int prio) {
  mysqlHandle * dbh = handle;
  MYSQL_RES * sql_res;
  MYSQL_ROW sql_row;
  char * escapedHash;
  char * scratch;
  int len;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  escapedHash = MALLOC(sizeof(HashCode160)*2+1);
  mysql_escape_string(escapedHash, 
		      (char *)query, 
		      sizeof(HashCode160));
  scratch = MALLOC(256);
  sprintf(scratch, 
	  "SELECT content,type,priority,doubleHash,fileOffset,"
	  "fileIndex FROM data%uof%u WHERE hash='%s'",
	  dbh->n,
          dbh->i,
	  escapedHash);
  mysql_query(dbh->dbf, 
	      scratch);
  if (mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR2: %s\n",
	mysql_error(dbh->dbf));
    FREE(escapedHash);
    FREE(scratch);
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }

  if (!(sql_res=mysql_store_result(dbh->dbf))) {
    LOG(LOG_ERROR, 
	"ERROR3: %s\n",
	mysql_error(dbh->dbf));
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    FREE(escapedHash);
    FREE(scratch);
    return SYSERR;
  }

  if (!(sql_row=mysql_fetch_row(sql_res))) {
    /* not error, just data not found */
    mysql_free_result(sql_res);
    FREE(escapedHash);
    FREE(scratch);
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }
    
  len = mysql_fetch_lengths(sql_res)[0];
  if (len > 0) {
    *result = MALLOC(len);
    memcpy(*result,
	   sql_row[0],
	   len);
  } else
    *result = NULL;

  ce->type = htons(atol(sql_row[1]));
  ce->importance = htonl(atol(sql_row[2]));
  if (ntohs(ce->type)==LOOKUP_TYPE_3HASH) {
    if (mysql_fetch_lengths(sql_res)[3] == sizeof(HashCode160))
      memcpy(&ce->hash, 
  	     sql_row[3],
	     sizeof(HashCode160));
  } else {
    memcpy(&ce->hash, 
    	   query,
	   sizeof(HashCode160));
  }
  ce->fileOffset = htonl(atol(sql_row[4]));
  ce->fileNameIndex = htons(atol(sql_row[5]));

  mysql_free_result(sql_res);

  if (prio != 0) {
    sprintf(scratch, 
	    "UPDATE data%uof%u SET priority=priority+%d WHERE hash='%s'",
	    dbh->n,
    	    dbh->i,
	    prio,
	    escapedHash);
    mysql_query(dbh->dbf,scratch); 
  }

  FREE(escapedHash);
  FREE(scratch);
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return len;
}

/**
 * Write content to the db.  Overwrites existing data.
 * If ce->type is LOOKUP_TYPE_3HASH, ce->hash will contain
 * a double hash which must be converted to 3HASH, later to be 
 * retrievable by 3HASH, but the 2HASH must be stored so it can
 * be retrieved by readContent(). For indexed content,
 * ce->fileOffset and ce->fileNameIndex must be stored.
 * Note that block can be NULL for on-demand encoded content
 * (in this case, len must also be 0).
 *
 * @param fn the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int writeContent(HighDBHandle handle,
		 ContentIndex * ce,
		 int len,
		 void * block) {
  mysqlHandle * dbh = handle;
  HashCode160 tripleHash;
  char * doubleHash;
  char * escapedBlock;
  char * escapedHash;
  char * scratch;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  
  escapedHash = MALLOC(2*sizeof(HashCode160)+1);
  
  if(ntohs(ce->type) == LOOKUP_TYPE_3HASH) {
    hash(&ce->hash, 
    	 sizeof(HashCode160),
	 &tripleHash);
    mysql_escape_string(escapedHash, 
			(char *)&tripleHash, 
			sizeof(HashCode160));
    doubleHash = MALLOC(2*sizeof(HashCode160)+1);
    mysql_escape_string(doubleHash, 
			(char *)&ce->hash, 
			sizeof(HashCode160));
  } else {
    doubleHash = NULL;
    mysql_escape_string(escapedHash, 
			(char *)&ce->hash, 
			sizeof(HashCode160));
  }
  
  escapedBlock = MALLOC(2*len+1);
  mysql_escape_string(escapedBlock, 
		      (char *)block, 
		      len);
  scratch = MALLOC(len*2+sizeof(HexName)*2+sizeof(HashCode160)*2+100+1);
  sprintf(scratch, 
	  "REPLACE INTO data%uof%u "
	  "(content,hash,priority,fileOffset,fileIndex,doubleHash,type)"
	  " VALUES ('%s','%s','%d','%d','%d','%s',%d)",
	  dbh->n,
	  dbh->i,
	  (len > 0 ? escapedBlock : ""),
	  escapedHash,
	  ntohl(ce->importance),
	  ntohl(ce->fileOffset),
	  ntohs(ce->fileNameIndex),
	  (doubleHash ? doubleHash : ""),
	  ntohs(ce->type));
  mysql_query(dbh->dbf, scratch);
  FREE(scratch);
  FREE(escapedBlock);
  FREE(escapedHash);
  FREENONNULL(doubleHash);
  if(mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR1: %s\n", 
	mysql_error(dbh->dbf));
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }
    
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return OK;
}

/**
 * Free space in the database by removing one file
 * @param name hashcode representing the name of the file (without directory)
 **/
int unlinkFromDB(HighDBHandle handle,
		 HashCode160 * name) {
  mysqlHandle * dbh = handle;
  char * escapedHash;
  char * scratch;

  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  escapedHash = MALLOC(2*sizeof(HashCode160)+1);
  mysql_escape_string(escapedHash, (char *)name, sizeof(HashCode160));
  scratch=MALLOC(sizeof(HashCode160)*2+100+1);
  sprintf(scratch, 
	  "DELETE FROM data%uof%u WHERE hash='%s'",	
	  dbh->n,
	  dbh->i,
  	  escapedHash);
  mysql_query(dbh->dbf,scratch);
  FREE(escapedHash);
  FREE(scratch);
  if(mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
	"ERROR: %s\n", mysql_error(dbh->dbf));
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  return OK;
}

/**
 * This function tries to efficiently get a random hash entry from 
 * the database by restricting the search with a randomized prefix.
 * The prefix length is started from the value where we can expect
 * to get 1 result. If we don't get that, the prefix length is
 * decremented, growing the expected number of results.
 *
 * FIXME: The way to calculate the initial prefix len might be borked
 **/
int getRandomContent(HighDBHandle handle,
		     ContentIndex * ce) {
  mysqlHandle * dbh = handle;
  MYSQL_RES * sql_res;
  MYSQL_ROW sql_row;
  char * escapedPrefix;
  char * prefix;
  char * scratch;
  int entries;
  int prefixlen;
  int found;
  int i;
  
  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  entries = countContentEntries(handle);
  prefixlen = (int)floor(log(entries)/(double)5.5452); /* 5.5452=log(256) */
  if (prefixlen>sizeof(HashCode160))
    prefixlen=sizeof(HashCode160);
  prefix = MALLOC(sizeof(HashCode160));
  escapedPrefix = MALLOC(2*sizeof(HashCode160)+1);
  scratch = MALLOC(2*sizeof(HashCode160)+256+1);

  found = NO;
  while((found == NO) && prefixlen>=0) {
    for (i=0;i<prefixlen;i++)
      prefix[i] = randomi(256);
    mysql_escape_string(escapedPrefix, prefix, prefixlen);

    sprintf(scratch, 
	    "SELECT hash,type,priority FROM data%uof%u "
	    "WHERE hash LIKE '%s%%' "
	    "AND type != %d "
	    "AND type != %d "
	    "ORDER BY RAND() LIMIT 1",
	    dbh->n,
    	    dbh->i,
	    escapedPrefix,
	    LOOKUP_TYPE_3HASH,
	    LOOKUP_TYPE_SUPER);
    mysql_query(dbh->dbf, scratch);
    if(mysql_error(dbh->dbf)[0]) {
      LOG(LOG_ERROR, 
  	  "ERROR: %s\n", 
	  mysql_error(dbh->dbf));
      FREE(prefix);
      FREE(scratch);
      FREE(escapedPrefix);
      MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
      return SYSERR;
    }
    if(!(sql_res=mysql_store_result(dbh->dbf))) {
      LOG(LOG_ERROR, 
  	  "ERROR: %s\n",
	  mysql_error(dbh->dbf));
      FREE(prefix);
      FREE(scratch);
      FREE(escapedPrefix);
      MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
      return SYSERR;
    }
    if(mysql_num_rows(sql_res)==0) {
      mysql_free_result(sql_res);
      prefixlen--;
      continue;
    }
    if(!(sql_row=mysql_fetch_row(sql_res))) {
      LOG(LOG_ERROR, 
  	  "ERROR: %s\n", 
	  mysql_error(dbh->dbf));
      FREE(prefix);
      FREE(scratch);
      FREE(escapedPrefix);
      mysql_free_result(sql_res);
      MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
      return SYSERR;
    }
    memcpy(&ce->hash,
    	   sql_row[0],
	   sizeof(HashCode160));
    ce->type = htons(atol(sql_row[1]));
    ce->importance = htonl(atol(sql_row[2]));
    found = YES;
    mysql_free_result(sql_res);
  }

  FREE(escapedPrefix);
  FREE(prefix);
  FREE(scratch);
 
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  if(found==YES) {
    return OK;
  } else {
    LOG(LOG_DEBUG, 
	"DEBUG: MySQL random didn't find anything!\n");
    return SYSERR;
  }
}


/**
 * Get the lowest priority of content in the store.
 **/
unsigned int getMinimumPriority(HighDBHandle handle) {
  mysqlHandle * dbh = handle;
  MYSQL_RES * sql_res;
  MYSQL_ROW sql_row;
  char * scratch;
  unsigned int minPrio = 0;
  
  MUTEX_LOCK(&dbh->DATABASE_Lock_);
  scratch = MALLOC(256);
 
  sprintf(scratch, 
	  "SELECT MIN(priority) FROM data%uof%u",
	  dbh->n,
	  dbh->i);
  mysql_query(dbh->dbf, scratch);
  FREE(scratch);
  if (!(sql_res=mysql_store_result(dbh->dbf))) {
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return 0;
  }

  if ((sql_row=mysql_fetch_row(sql_res))) {
    minPrio = atol(sql_row[0]);
  }
  mysql_free_result(sql_res);
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);

  return minPrio;
}


/**
 * Deletes some least important content
 * 
 * @param count the number of entries to delete
 **/
int deleteContent(HighDBHandle handle,
		  int count,
		  EntryCallback callback,
		  void * closure) {
  mysqlHandle * dbh = handle;
  MYSQL_RES * sql_res;
  MYSQL_ROW sql_row;
  HashCode160 * deleteThese;
  char * escapedHash;
  char * scratch;
  int i=0;
  
  MUTEX_LOCK(&dbh->DATABASE_Lock_);

  scratch = MALLOC(256);
 
  /* Collect hashes to delete */
  sprintf(scratch, 
	  "SELECT hash FROM data%uof%u "
	  "ORDER BY priority LIMIT %d",
	  dbh->n,
	  dbh->i,
  	  count);
  mysql_query(dbh->dbf, scratch);
  if (mysql_error(dbh->dbf)[0]) {
    LOG(LOG_ERROR, 
        "ERROR: %s\n",
	mysql_error(dbh->dbf));
    FREE(scratch);
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }
  if (!(sql_res=mysql_use_result(dbh->dbf))) {
    FREE(scratch);
    LOG(LOG_ERROR, 
        "ERROR: %s\n", 
	mysql_error(dbh->dbf));
    MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
    return SYSERR;
  }
  
  deleteThese = MALLOC(count*sizeof(HashCode160));
  while ((sql_row=mysql_fetch_row(sql_res))) {
    memcpy(&deleteThese[i++],
           sql_row[0],
	   sizeof(HashCode160));
  }
  mysql_free_result(sql_res);
 
  /* Delete collected hashes */
  count=i;
  escapedHash = MALLOC(2*sizeof(HashCode160)+1);
  for(i=0;i<count;i++) {    
    ContentIndex ce;
    void * data;
    int dlen;
    
    data = NULL;
    dlen = readContent(handle,
		       &deleteThese[i],
		       &ce,
		       &data,
		       0);
    if (dlen >= 0) 
      if (callback != NULL)
	callback(&deleteThese[i],
		 &ce,
		 data,
		 dlen,
		 closure);
    FREENONNULL(data);

    mysql_escape_string(escapedHash, 	
                        (char *)&deleteThese[i],
			sizeof(HashCode160));
    sprintf(scratch, 
	    "DELETE FROM data%uof%u WHERE hash='%s'",
	    dbh->n,
	    dbh->i,
    	    escapedHash);
    mysql_query(dbh->dbf,scratch);
    if(mysql_error(dbh->dbf)[0])
      LOG(LOG_ERROR, 
          "ERROR: %s\n",
	  mysql_error(dbh->dbf));
  }
    
  FREE(escapedHash);  
  FREE(scratch);
  FREE(deleteThese);
  
  MUTEX_UNLOCK(&dbh->DATABASE_Lock_);
  
  return OK;
}


/**
 * Estimate how many blocks can be stored in the DB
 * before the quota is reached.
 *
 * @quota the number of kb available for the DB
 **/ 
int estimateAvailableBlocks(HighDBHandle handle,
			    int quota) {
  return quota - ( ( ( countContentEntries(handle) * 120) / 100)); /* according to EH, 100 MB use 113 MB on the drive, 
								      a little over-estimation is probably better... */
}


/**
 * Close and delete the database.
 **/
void deleteDatabase(HighDBHandle handle) {
  mysqlHandle * dbh = handle;
  char * scratch;
  
  MUTEX_DESTROY(&dbh->DATABASE_Lock_);
  scratch = MALLOC(128);
  sprintf(scratch,
  	  "DROP TABLE data%uof%u", 
	  dbh->n,
	  dbh->i);
  mysql_query(dbh->dbf,
  	      scratch);
  FREE(scratch);
  mysql_close(dbh->dbf);
  FREE(dbh);
}

/* end of high_mysql.c */
