/*
     This file is part of GNUnet.
     (C) 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/encoding/deletefile.c
 * @author Krista Bennett
 * @author Christian Grothoff
 *
 * Break file that is deleted into blocks and encrypts
 * them according to the CHK-triple-hash-tree scheme.
 * Then sends delete-requests to gnunetd.
 **/

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "util/checksum.h"
#include "util/configuration.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/storage.h"
#include "util/tcp_return.h"
#include "util/xmalloc.h"
#include "applications/afs/encoding/deleteutil.h"
#include "applications/afs/util/contentencoding.h"

/**
 * Ask gnunetd to append a filename,
 * @return the index, -1 on error
 **/
static int askDeleteFilename(GNUNET_TCP_SOCKET * sock,
			     char * filename) {
  AFS_CS_INDEX_FILE * request;
  int result;
  
  filename 
    = expandFileName(filename);   
  request = MALLOC(sizeof(AFS_CS_INDEX_FILE));
  request->header.size 
    = htons(sizeof(AFS_CS_INDEX_FILE));
  request->header.tcpType 
    = htons(AFS_CS_PROTO_UNINDEX_FILE);
  memset(&request->filename[0], 
	 0,
	 CS_FILE_LIST_FILENAME);
  memcpy(&request->filename[0],
	 filename,
	 strlen(filename));
  FREE(filename);
  if (SYSERR == writeToSocket(sock,
			      &request->header)) {
    FREE(request);
    LOG(LOG_WARNING, 
	"WARNING: could not send data to gnunetd. "\
	"Is gnunetd running?\n");
    return -1;
  } 
  FREE(request);
  if (SYSERR == readTCPResult(sock,
			      &result)) {
    LOG(LOG_WARNING, 
	"WARNING: could not receive data from gnunetd. Is gnunetd running?\n");
    result = -1;
  }
  return result;
}
  
/**
 * De-facto main method. Deletes a file under the given
 * name into the local GNUnet node.
 *
 * @param filename the name of the file to delete
 * @param model the delete model used to
 *        update status information; points to NULL if
 *        no status updates shall be given, otherwise 
 *        to a method that takes two size_t arguments
 *        (retrieved so far, total).
 * @param model_data pointer that is passed to the model method
 * @return SYSERR on error, OK on success
 **/
int deleteFile(GNUNET_TCP_SOCKET * sock,
	       char * filename, 
	       ProgressModel model,
	       void * model_data) {
  NodeContext nc;
  size_t filesize;
  Block * top;

  filename = expandFileName(filename);
  filesize = (size_t) getFileSize(filename);
  nc.pmodel = model;
  nc.data = model_data;
  memset(&nc.stats, 0, sizeof(ProgressStats));
  nc.stats.filesize = filesize;
  nc.priority = 0;
  nc.index = askDeleteFilename(sock, filename);
  if (SYSERR == createIOContext(&nc.ioc,
				filesize,
				filename,
				YES)) {    
    FREE(filename);
    return SYSERR;
  }
  if (filesize <= sizeof(CONTENT_Block))
    top = createTopDBlock(filesize);
  else
    top = createTopIBlock(filesize);
  if (SYSERR == top->delete(top, &nc, sock)) {
    top->done(top, NULL);
    freeIOC(&nc.ioc, NO);
    FREE(filename);
    return SYSERR;
  }
  freeIOC(&nc.ioc, NO);
  FREE(filename);
  top->done(top, NULL);
  return OK;
}


/* end of deleteutil.c */
