/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/download.c
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 * This is the code for handling download and other related
 * windows for the gtk+ client.
 **/

#include "config.h"

#include <errno.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>

#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/xmalloc.h"
#include "applications/afs/encoding/downloadutil.h"
#include "applications/afs/encoding/requestmanager.h"
#include "applications/afs/gtkui/helper.h"
#include "applications/afs/gtkui/download.h"
#include "applications/afs/gtkui/main.h"


/* Suitable values of DownloadModel.downloadStatus */
#define DOWNLOAD_COMPLETE 0
#define DOWNLOAD_FAILED   1
#define DOWNLOAD_ABORTED  2
#define DOWNLOAD_PENDING  3

typedef struct {
  RootNode root;
  char * fileName;
  pthread_t downloadThread;
  RequestManager * rm;
  Semaphore * doneSem;
  cron_t downloadStartTime;
  int downloadStatus;
  GtkWidget * progressBar;
  GtkWidget * progressBarWindow;
  GtkWidget * progressBarWindowBox;
  GtkWidget * progressRetries;
} DownloadModel;


/**
 * "delete_event" handler for the progressBarWindow. Sets the 
 * download state as aborted, and releases the wait semaphore for 
 * the thread awaiting download completion. Returns TRUE so that 
 * gtk doesn't destroy the window (downloadFile_ will!).
 *
 * @param widget here progressBarWindow
 * @param dlm the accessor to the GTK window
 **/
static gint abortDownload(GtkWidget * widget,
			  GdkEvent * event,
			  DownloadModel * dlm) {
  LOG(LOG_DEBUG,
      "DEBUG: abortDownload %x (%x)\n",
      dlm, widget);
  dlm->downloadStatus = DOWNLOAD_ABORTED;
  SEMAPHORE_UP(dlm->doneSem);
  return TRUE;
}

/**
 * "YAY/BAH" button handler for the progressBarWindow.
 *
 * @param widget here progressBarWindow
 * @param dlm the accessor to the GTK window
 **/
static void abortDownloadButton(GtkWidget * widget,
				DownloadModel * dlm) {
  LOG(LOG_DEBUG,
      "DEBUG: abortDownloadButton %x (%x)\n",
      dlm, widget);
  SEMAPHORE_UP(dlm->doneSem);
}

/**
 * This method is called by the download code to notify the user
 * interface of the download progress.
 *
 * @param stats the new statistical values
 * @param dlm the accessor to the GTK window
 **/
static void modelCallback(ProgressStats * stats,
			  DownloadModel * dlm) {
  double currentRetryAvg, averageBps;
  char scratch[256];

  if (dlm->downloadStatus != DOWNLOAD_PENDING) 
    return;
  if (stats->filesize == stats->progress) {
    if (stats->filesize == 0)       
      dlm->downloadStatus = DOWNLOAD_FAILED;
    else
      dlm->downloadStatus = DOWNLOAD_COMPLETE;
    SEMAPHORE_UP(dlm->doneSem); /* signal: we're done with download */
    return;
  }
  
  /*
  LOG(LOG_DEBUG, 
      "DEBUG: progress %d size %d pend %d sent %d"
      " currentr %d totalr %d ttl %d estimate %d\n",
      stats->progress, 
      stats->filesize,
      stats->requestsPending, 
      stats->requestsSent, 
      stats->currentRetries,
      stats->totalRetries,
      stats->currentTTL, 
      stats->duplicationEstimate);
  */

  if(stats->requestsSent > 0)
    currentRetryAvg = 
      (double)stats->currentRetries / 
      (double)stats->requestsSent;
  else
    currentRetryAvg = 0;
    
  if (cronTime(NULL) - dlm->downloadStartTime > 0)
    averageBps =
      (double)(stats->progress) / 
      (((double)(cronTime(NULL) - dlm->downloadStartTime))/(double)cronSECONDS);
  else
    averageBps = 0;
 
  if (stats->progress == 0) {
    sprintf(scratch, 
 	    "ActiveReq: %4d AvgRetries/Active: %3.1f\nTotalRetries: %d\nFetching metablocks", 
	    stats->requestsSent,
  	    currentRetryAvg,
	    stats->totalRetries);
  } else {
    sprintf(scratch, 
 	    "ActiveReq: %4d AvgRetries/Active: %3.1f\nTotalRetries: %d\nAvgBPS: %8.3f", 
	    stats->requestsSent,
  	    currentRetryAvg,
	    stats->totalRetries,
	    averageBps);
  }

  /* update view */
  gdk_threads_enter();
  gtk_progress_set_value(GTK_PROGRESS(dlm->progressBar),
			 stats->progress);
  gtk_label_set_text(GTK_LABEL(dlm->progressRetries), 
		     scratch);
  gdk_threads_leave();
}

/**
 * Main function of the download thread. This function terminates
 * automagically once the download is complete or if the download
 * window is closed.  It is responsible for stopping the
 * requestmanager.
 * 
 * @param dlm the download model
 **/
static void downloadFile_(DownloadModel * dlm) {
  GtkWidget * button;
  
  LOG(LOG_DEBUG,
      "DEBUG: Entering downloadFile_ for %s (%x)\n", 
      dlm->fileName,
      dlm);

  /* initiate download */
  cronTime(&dlm->downloadStartTime);
  
  /* this starts the "real" download thread in the background,
     with "modelCallback" called back to tell us about the progress */
  dlm->rm = downloadFile(&dlm->root.header.fileIdentifier.chk,
		         dlm->root.header.fileIdentifier.crc,
			 dlm->root.header.fileIdentifier.file_length,
			 dlm->fileName,
			 (ProgressModel)&modelCallback,
			 dlm);

  /* Wait here until download is complete or
     the window is closed or gnunet-gtk is terminated */
  LOG(LOG_DEBUG,
      "DEBUG: waiting for DL completion (%x)\n",
      dlm);
  SEMAPHORE_DOWN(dlm->doneSem);
  LOG(LOG_DEBUG,
      "DEBUG: download complete (%d) enter destroyRequestManager (%x)\n",
      dlm->downloadStatus,
      dlm);

  /* stop the RequestManager */
  destroyRequestManager(dlm->rm); 

  /*
    ok, now why are we here? 4 possibilities:
     a) download aborted (user closed window)
     b) gnunet-gtk terminated (same as download aborted)
     c) download failed (gnunetd exit, out-of-space)
     d) download completed 

     In case "d" we show the "YAY" window
     and wait for another signal.
     In case "c" we show the "BAH" window
     and wait for another signal.
  */
  switch (dlm->downloadStatus) {
  case DOWNLOAD_COMPLETE:
    gdk_threads_enter();  
    gtk_progress_set_format_string(GTK_PROGRESS(dlm->progressBar),
				   "Download complete");
    button = gtk_button_new_with_label("Yay!");
    gtk_signal_connect(GTK_OBJECT(button), 
		       "clicked",
		       GTK_SIGNAL_FUNC(abortDownloadButton),
		       dlm);
    gtk_box_pack_start(GTK_BOX(dlm->progressBarWindowBox), 
		       button, 
		       FALSE, 
		       FALSE, 
		       0);
    gtk_widget_show(button);  
    gdk_threads_leave();
    SEMAPHORE_DOWN(dlm->doneSem); /* wait for window closing */
    break;
  case DOWNLOAD_FAILED:
    gdk_threads_enter();  
    gtk_progress_set_format_string(GTK_PROGRESS(dlm->progressBar),
		    	 	  "Download failed!");
    button = gtk_button_new_with_label("Bah.");
    gtk_signal_connect(GTK_OBJECT(button), 
		       "clicked",
		       GTK_SIGNAL_FUNC(abortDownloadButton),
		       dlm);
    gtk_box_pack_start(GTK_BOX(dlm->progressBarWindowBox), 
		       button, 
		       FALSE, 
		       FALSE, 
		       0);
    gtk_widget_show(button);  
    gdk_threads_leave();
    SEMAPHORE_DOWN(dlm->doneSem); /* wait for window closing */    
    break;
  default:
    /* do nothing */
    break;
  }

  /* finally, close the window and free dlm resources */
  gdk_threads_enter();  
  destroyWidget(NULL, dlm->progressBarWindow);
  gdk_threads_leave();
  SEMAPHORE_FREE(dlm->doneSem);
  FREE(dlm->fileName);
  FREE(dlm);
}

/**
 * Open the download window and start the download of a file in the
 * background. The method executes during a signal handler, so a GTK
 * lock is not required to to GUI operations.
 *
 * @param filename the name of the file to download
 * (must copy, will be freed by caller)
 * @param dlm some information about what to download 
 * (will be freed by caller, must copy!)
 **/
void startDownload(gchar * filename,
		   RootNode * root) {
  GtkWidget * window;
  GtkWidget * box;
  GtkWidget * label;
  GtkObject * adjustment;
  GtkWidget * progressBar;
  char * fileNameRoot;
  int i;
  DownloadModel * dlm;

  dlm = MALLOC(sizeof(DownloadModel));
  memset(dlm, 
	 0, 
	 sizeof(DownloadModel));
  dlm->fileName = STRDUP(filename);
  memcpy(&dlm->root,
	 root,
	 sizeof(RootNode));

  /* create download window */
  fileNameRoot = dlm->fileName;
  for (i=strlen(dlm->fileName)-1;i>=0;i--) {
    if (dlm->fileName[i] == '/') {
      fileNameRoot = &dlm->fileName[i+1];
      break;
    }
  }  
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), 
		       fileNameRoot);

  /* hook up signal for window closing */
  gtk_signal_connect(GTK_OBJECT(window), 
		     "delete_event",
		     GTK_SIGNAL_FUNC(abortDownload), 
		     dlm);

  /* create a box that will contain the progress bar
     and the status text */
  box = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), 
		    box);
  gtk_container_set_border_width(GTK_CONTAINER(window), 
				 10);

  /* add a progressbar */
  progressBar = gtk_progress_bar_new();
  gtk_progress_set_show_text(GTK_PROGRESS(progressBar),
			     1);
  gtk_progress_set_format_string(GTK_PROGRESS(progressBar),
				 (gchar*)"%v bytes downloaded");
  adjustment = gtk_adjustment_new(0,
				  0,
				  dlm->root.header.fileIdentifier.file_length,
				  1,
				  0,
				  0);
  gtk_progress_set_adjustment(GTK_PROGRESS(progressBar),
			      GTK_ADJUSTMENT(adjustment));
  gtk_box_pack_start(GTK_BOX(box), 
		     progressBar, 
		     TRUE, 
		     TRUE,
		     0);
  gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(progressBar),
				   GTK_PROGRESS_LEFT_TO_RIGHT);

  /* add label (empty for now, filled with detailed progress info later) */
  label = gtk_label_new(" ");
  gtk_box_pack_start(GTK_BOX(box),
		     label, 
		     FALSE, 
		     FALSE, 
		     0);

  /* remember all of these for updates in the DLM struct */
  dlm->progressBar = progressBar;
  dlm->progressBarWindow = window;
  dlm->progressBarWindowBox = box;
  dlm->progressRetries = label;  
  dlm->downloadStatus = DOWNLOAD_PENDING;
  dlm->doneSem = SEMAPHORE_NEW(0);
  
  /* display download window */
  gtk_widget_show(label);
  gtk_widget_show(progressBar);
  gtk_widget_show(box);
  gtk_widget_show(window);

  /* create thread that runs the download */
  if (0 != PTHREAD_CREATE(&dlm->downloadThread,
			  (void * (*)(void *)) &downloadFile_,
			  dlm,
			  64 * 1024))
    errexit("FATAL: could not create download thread (%s)!\n",
	    strerror(errno));
  pthread_detach(dlm->downloadThread);  
}

/* end of download.c */
