/**
 * @file test/cry_comp.c
 * @brief test compatibility between libgcrypt and OpenSSL
 * @author Christian Grothoff
 **/

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <gcrypt.h>
#include "util/hashing.h"
#include <netinet/in.h>
#include "util/logging.h"
#include "util/xmalloc.h"
#include "util/returnvalues.h"
#include "hk_crypto.h"
#include "hk_gcrypt.h"

#define TESTSTRING "Hello World\0"
#define MAX 20
#define ITER 100

static int testEncryptDecryptO2G() {
  RSA * hostkey;  
  PublicKey pkey;
  RSAEncryptedData target;
  char result[MAX];
  int i;
  time_t start;


  hostkey = omakeHostkey();
  ogetPublicKey(hostkey, &pkey);

  time(&start);
  for (i=0;i<ITER;i++) {
    fprintf(stderr, ".");
    if (SYSERR == gencryptHostkey(TESTSTRING,
				  strlen(TESTSTRING)+1,
				  &pkey,
				  &target)) {
      printf("testEncryptDecryptO2G failed (encrypt failed)!\n");
      return SYSERR;
    }
    if (-1 == odecryptHostkey(hostkey, &target, result, MAX)) {
      printf("testEncryptDecryptO2G failed (decrypt failed)!\n");
      return SYSERR;
    }
    if (strncmp(TESTSTRING, result,
		strlen(TESTSTRING)) != 0) {
      printf("testEncryptDecryptO2G failed (invalid result)!\n");
      return SYSERR;
    }
  }
  printf("%d RSA encrypt/decrypt operations O2G %ds\n", 
	 ITER,
	 (int) (time(NULL)-start));
  ofreeHostkey(hostkey);
  return OK;
}

static int testEncryptDecryptG2O() {
  GCRY_SEXP hostkey;
  PublicKey pkey;
  RSAEncryptedData target;
  char result[MAX];
  int i;
  time_t start;


  hostkey = gmakeHostkey();
  ggetPublicKey(hostkey, &pkey);

  time(&start);
  for (i=0;i<ITER;i++) {
    fprintf(stderr, ".");
    if (SYSERR == oencryptHostkey(TESTSTRING,strlen(TESTSTRING)+1,&pkey,
				  &target)) {
      printf("testEncryptDecryptG2O failed (encrypt failed)!\n");
      return SYSERR;
    }
    if (-1 == gdecryptHostkey(hostkey, &target, result, MAX)) {
      printf("testEncryptDecryptG2O failed (decrypt failed)!\n");
      return SYSERR;
    }
    if (strncmp(TESTSTRING, result,
		strlen(TESTSTRING)) != 0) {
      printf("testEncryptDecryptG2O failed (invalid result)!\n");
      return SYSERR;
    }
  }
  printf("%d RSA encrypt/decrypt operations G2O %ds\n", 
	 ITER,
	 (int) (time(NULL)-start));
  gfreeHostkey(hostkey);
  return OK;
}

static int testSignVerifyO2G() {
  RSA * hostkey;
  Signature sig;
  PublicKey pkey;
  int i;
  time_t start;
  
  hostkey = omakeHostkey();
  ogetPublicKey(hostkey, &pkey);
  time(&start);
  for (i=0;i<ITER;i++) {
    fprintf(stderr, ".");
    if (SYSERR == osign(hostkey, strlen(TESTSTRING), TESTSTRING, &sig)) {
      printf("testSignVerifyO2G failed (sign)!\n");
      return SYSERR;
    }
    if (SYSERR == gverifySig(TESTSTRING, strlen(TESTSTRING), &sig, &pkey)) {
      printf("testSignVerifyO2G failed!\n");
      return SYSERR;
    }
  }
  printf("%d RSA sign/verify operations O2G %ds\n", 
	 ITER, 
	 (int) (time(NULL)-start));
  ofreeHostkey(hostkey);
  return OK;
}

static int testSignVerifyG2O() {
  GCRY_SEXP hostkey;
  Signature sig;
  PublicKey pkey;
  int i;
  time_t start;
  
  hostkey = gmakeHostkey();
  ggetPublicKey(hostkey, &pkey);
  time(&start);
  for (i=0;i<ITER;i++) {
    fprintf(stderr, ".");
    if (SYSERR == gsign(hostkey, strlen(TESTSTRING), TESTSTRING, &sig)) {
      printf("testSignVerifyG2O failed (sign)!\n");
      return SYSERR;
    }
    if (SYSERR == overifySig(TESTSTRING, strlen(TESTSTRING), &sig, &pkey)) {
      printf("testSignVerifyG2O failed!\n");
      return SYSERR;
    }
  }
  printf("%d RSA sign/verify operations G2O %ds\n", 
	 ITER, 
	 (int) (time(NULL)-start));
  gfreeHostkey(hostkey);
  return OK;
}

static int testHostkeyEncodingO2G() {
  RSA * hostkey;
  GCRY_SEXP hostkey2;
  HostKeyEncoded * encoding;
  PublicKey pkey;
  RSAEncryptedData target;
  char result[MAX];
  int i;
  time_t start;

  hostkey = omakeHostkey();

  time(&start);
  for (i=0;i<ITER;i++) {
    fprintf(stderr, ".");
    ogetPublicKey(hostkey, &pkey);
    if (SYSERR == oencryptHostkey(TESTSTRING,
				 strlen(TESTSTRING)+1,
				 &pkey,
				  &target)) {
      printf("testEncodeEncryptDecodeDecryptO2G failed (encrypt)!\n");
      return SYSERR;
    }
    encoding = oencodeHostkey(hostkey);
    if (encoding == NULL) {
      printf("testEncodeEncryptDecodeDecryptO2G failed (encode)!\n");
      return SYSERR;
    }
    hostkey2 = gdecodeHostkey(encoding);
    free(encoding);
    if (-1 == gdecryptHostkey(hostkey2, &target, result, MAX)) {
      printf("testEncodeEncryptDecodeDecryptO2G failed (decrypt)!\n");
      return SYSERR;
    }  
    if (strncmp(TESTSTRING, result,
		strlen(TESTSTRING)) != 0) {
      printf("testEncodeEncryptDecodeDecryptO2G failed (result invalid!\n");
      return SYSERR;
    }
  }  
  gfreeHostkey(hostkey2);
  ofreeHostkey(hostkey);
  printf("%d RSA encrypt/encode/decode/decrypt operations %ds\n", 
	 ITER, 
	 (int) (time(NULL)-start));
  return OK;
}

static int testHostkeyEncodingG2O() {
  GCRY_SEXP hostkey;
  RSA * hostkey2;
  HostKeyEncoded * encoding;
  PublicKey pkey;
  RSAEncryptedData target;
  char result[MAX];
  int i;
  time_t start;

  hostkey = gmakeHostkey();

  time(&start);
  for (i=0;i<ITER;i++) {
    fprintf(stderr, ".");
    ggetPublicKey(hostkey, &pkey);
    if (SYSERR == gencryptHostkey(TESTSTRING,
				 strlen(TESTSTRING)+1,
				 &pkey,
				  &target)) {
      printf("testEncodeEncryptDecodeDecryptG2O failed (encrypt)!\n");
      return SYSERR;
    }
    encoding = gencodeHostkey(hostkey);
    if (encoding == NULL) {
      printf("testEncodeEncryptDecodeDecryptG2O failed (encode)!\n");
      return SYSERR;
    }
    hostkey2 = odecodeHostkey(encoding);
    free(encoding);
    if (-1 == odecryptHostkey(hostkey2, &target, result, MAX)) {
      printf("testEncodeEncryptDecodeDecryptG2O failed (decrypt)!\n");
      return SYSERR;
    }  
    if (strncmp(TESTSTRING, result,
		strlen(TESTSTRING)) != 0) {
      printf("testEncodeEncryptDecodeDecryptG2O failed (result invalid)!\n");
      return SYSERR;
    }
  }  
  gfreeHostkey(hostkey);
  ofreeHostkey(hostkey2);
  printf("%d RSA encrypt/encode/decode/decrypt operations %ds\n", 
	 ITER, 
	 (int) (time(NULL)-start));
  return OK;
}


int main(int argc, char * argv[]) {
  int failureCount = 0;

  oinitRAND();
  ginitRAND();
  gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);
  /* No valuable keys are create, so we can speed up our RNG. */
  gcry_control (GCRYCTL_ENABLE_QUICK_RANDOM, 0);

  printf("Test 1\n");
  if (OK != testEncryptDecryptO2G())
     failureCount++;
  printf("Test 2\n");
  if (OK != testSignVerifyO2G())
    failureCount++;     
  printf("Test 3\n");
  if (OK != testHostkeyEncodingO2G()) 
    failureCount++;
  printf("Test 4\n");
  if (OK != testEncryptDecryptG2O())
     failureCount++;
  printf("Test 5\n");
  if (OK != testSignVerifyG2O())
    failureCount++;     
  printf("Test 6\n");
  if (OK != testHostkeyEncodingG2O()) 
    failureCount++;


  if (failureCount == 0)
    return 0;
  else {
    printf("\n\n%d TESTS FAILED!\n\n",
	   failureCount);
    return -1;
  }
}
